
Require Import List.

(* Syntax of F< types *)

Parameter Var: Set.

Inductive Tp: Set := top: Tp
                  |  var: Var -> Tp
                  |  arr: Tp -> Tp -> Tp
                  |  fa : Tp -> (Var -> Tp) -> Tp.

Coercion var: Var >-> Tp.

(* Typing Environment and its Well-fordmedness *)

Definition envTp: Set := (list (Var * Tp)).

Inductive Gfresh (x:Var): envTp -> Prop :=
          GfVoid: Gfresh x nil
        | GfGrow: forall G:envTp, forall y:Var, forall T:Tp,
                  Gfresh x G -> ~x=y ->
                  Gfresh x (cons (y,T) G).

Inductive isin (x:Var): Tp -> Prop :=
          isin_var: isin x x
        | isin_arr: forall S T:Tp, 
                    isin x S \/ isin x T ->
                    isin x (arr S T)
        | isin_fa : forall S:Tp, forall T:Var->Tp,
                    isin x S \/ (forall y:Var, ~x=y -> isin x (T y)) ->
                    isin x (fa S T).

Inductive isinG (x:Var) (T:Tp): envTp -> Prop :=
          checkG: forall G:envTp, forall y:Var, forall U:Tp,
                  (x=y /\ T=U) \/ isinG x T G ->
                  isinG x T (cons (y,U) G).

Definition Gclosed (T:Tp) (G:envTp): Prop :=
           forall x:Var,
           (isin x T) -> exists U:Tp, isinG x U G.

Inductive notin (x:Var): Tp -> Prop :=
          notin_top: notin x top
        | notin_var: forall y:Var, 
                     ~x=y ->
                     notin x y
        | notin_arr: forall s t:Tp, 
                     notin x s -> notin x t ->
                     notin x (arr s t)
        | notin_fa : forall s:Tp, forall t:Var->Tp,
                     (notin x s) -> 
                     (forall y:Var, ~x=y -> notin x (t y)) ->
                     notin x (fa s t).

Inductive okEnv: envTp -> Prop :=
          okVoid: okEnv nil
        | okGrow: forall G:envTp, forall x:Var, forall T:Tp,
                  okEnv G ->
                  Gfresh x G -> Gclosed T G ->
                  okEnv (cons (x,T) G).

(* Subtyping *)   

Inductive subTp: envTp -> Tp -> Tp -> Prop :=

          sub_top: forall G: envTp, forall S: Tp,
                   okEnv G -> Gclosed S G ->
                   subTp G S top

        | sub_var: forall G: envTp, forall x: Var, forall U: Tp,
                   okEnv G -> isinG x U G ->
                   subTp G x x

        | sub_trs: forall G: envTp, forall x: Var, forall U T: Tp,
                   isinG x U G -> subTp G U T ->
                   subTp G x T

        | sub_arr: forall G: envTp, forall S1 S2 T1 T2: Tp,
                   subTp G T1 S1 -> subTp G S2 T2 ->
                   subTp G (arr S1 S2) (arr T1 T2)

        | sub_fa : forall G: envTp,
                   forall S1 T1: Tp, forall S2 T2: Var->Tp,
                   subTp G T1 S1 ->
                   (forall x:Var, okEnv (cons (x,T1) G) ->
                                  subTp (cons (x,T1) G) (S2 x) (T2 x)) ->
                   subTp G (fa S1 S2) (fa T1 T2).